


from readcfg import CFG_Vertex, CFG_Edge
import ilp, readcfg
import math

EXACT_FREQUENCIES = False

def Name(edge):
    assert isinstance(edge, CFG_Edge)
    return "E%06x" % edge.id

def Supername(edge, index):
    return "%sI%02x" % (Name(edge), index)

def Flow_Is_Conserved(edges):
    if ( len(edges) == 0 ):
        # entrance/exit - flow is 1
        return "1"

    equation = []
    for e in edges:
        equation.append(Name(e))
    return '+'.join(equation) 

def Common_Setup(cfg_entrance, model):
    assert isinstance(cfg_entrance, CFG_Vertex)
    assert isinstance(model, ilp.ILP_Solver)

    model.Reset_All()
    (vertexes, edges) = readcfg.Find_V_E(cfg_entrance)
    edge_dict = dict()

    for e in edges:
        name = Name(e)
        model.Add_Var(name, ilp.VAR_INTEGER)
        edge_dict[ name ] = e

        if EXACT_FREQUENCIES:
            model.Add_Term(name, '<=', str(e.max_frequency))
        elif e.loop_header != None:
            model.Add_Term(Name(e.loop_header), '<=', 
                        str(e.loop_header.max_frequency))
            model.Add_Term(name, '<=', str(e.max_frequency))
            fdiff = 1 + int(math.floor(e.max_frequency /
                        e.loop_header.max_frequency))
            model.Add_Term(name, '<=', '%s * %u' % (
                        Name(e.loop_header), fdiff))

    for v in vertexes:
        lhs = Flow_Is_Conserved(v.entrances)
        rhs = Flow_Is_Conserved(v.exits)
        model.Add_Term(lhs, '=', rhs)

    model.Add_Var("WCET", ilp.VAR_MIXED)
    model.Add_Problem("WCET")
    return (vertexes, edges, edge_dict)

def Common_Solver(model, edge_dict, ur_fn):
    def Results(key, value):
        if ( key == "WCET" ):
            if (value <= 0):
                raise ilp.ILPError("WCET <= 0")
        elif ( edge_dict.has_key(key) ):
            e = edge_dict[ key ]
            if EXACT_FREQUENCIES:
                assert value <= e.max_frequency, (
                    "%s --> %u <= %u" % (key,
                            value, e.max_frequency))
                
            e.frequency = value
        else:
            assert ur_fn != None
            ur_fn(key, value)

    model.Solve(Results)

    # WCET recomputed in case of floating-point imprecision
    # (Z has to be a "mixed" variable because of limited range
    # of 32-bit integers, which the x86_64 version of glpsol
    # insists on using for some reason.)

    z = 0
    for (key, edge) in edge_dict.iteritems():
        z += edge.frequency * edge.cost

    if (z <= 0):
        raise ilp.ILPError("WCET <= 0")
    return z


def Simple_IPET(cfg_entrance, model):
    (vertexes, edges, edge_dict) = Common_Setup(cfg_entrance, model)

    wcet = []
    for e in edges:
        assert e.cost >= 0
        if ( e.cost != 0 ):
            wcet.append('%d*%s' % (e.cost, Name(e)))

    model.Add_Term("WCET", "=", '+'.join(wcet))

    return Common_Solver(model, edge_dict, None)

